@@ -1,13 +1,11 @@
-from .codeact_agent import CodeActAgent
-from .prompts import (
-    DEFAULT_CODE_ACT_SYSTEM_PROMPT,
+from droidrun.agent.codeact.codeact_agent import CodeActAgent
+from droidrun.agent.codeact.prompts import (
     DEFAULT_CODE_ACT_USER_PROMPT,
     DEFAULT_NO_THOUGHTS_PROMPT
 )
 
 __all__ = [
     "CodeActAgent",
-    "DEFAULT_CODE_ACT_SYSTEM_PROMPT",
     "DEFAULT_CODE_ACT_USER_PROMPT",
     "DEFAULT_NO_THOUGHTS_PROMPT"
 ]
