@@ -18,7 +18,9 @@ from droidrun.agent.codeact.events import (
     TaskThinkingEvent,
     EpisodicMemoryEvent,
 )
-from droidrun.agent.common.events import ScreenshotEvent
+from droidrun.agent.common.constants import LLM_HISTORY_LIMIT
+from droidrun.agent.common.events import RecordUIStateEvent, ScreenshotEvent
+from droidrun.agent.usage import get_usage_from_response
 from droidrun.agent.utils import chat_utils
 from droidrun.agent.utils.executer import SimpleCodeExecutor
 from droidrun.agent.codeact.prompts import (
@@ -108,7 +110,8 @@ class CodeActAgent(Workflow):
         """Prepare chat history from user input."""
         logger.info("💬 Preparing chat for task execution...")
 
-        self.chat_memory: Memory = await ctx.get(
+
+        self.chat_memory: Memory = await ctx.store.get(
             "chat_memory", default=Memory.from_defaults()
         )
 
@@ -134,7 +137,7 @@ class CodeActAgent(Workflow):
 
         await self.chat_memory.aput(self.user_message)
 
-        await ctx.set("chat_memory", self.chat_memory)
+        await ctx.store.set("chat_memory", self.chat_memory)
         input_messages = self.chat_memory.get_all()
         return TaskInputEvent(input=input_messages)
 
@@ -161,25 +164,28 @@ class CodeActAgent(Workflow):
         model = self.llm.class_name()
         
         if "remember" in self.tool_list and self.remembered_info:
-            await ctx.set("remembered_info", self.remembered_info)
+            await ctx.store.set("remembered_info", self.remembered_info)
             chat_history = await chat_utils.add_memory_block(self.remembered_info, chat_history)
 
         for context in self.required_context:
-            if model == "DeepSeek":
-                logger.warning(
-                    "[yellow]DeepSeek doesnt support images. Disabling screenshots[/]"
-                )
-            elif self.vision == True and context == "screenshot":
+            if context == "screenshot":
+                # if vision is disabled, screenshot should save to trajectory
                 screenshot = (self.tools.take_screenshot())[1]
                 ctx.write_event_to_stream(ScreenshotEvent(screenshot=screenshot))
 
-                await ctx.set("screenshot", screenshot)
-                chat_history = await chat_utils.add_screenshot_image_block(screenshot, chat_history)
+                await ctx.store.set("screenshot", screenshot)
+                if model == "DeepSeek":
+                    logger.warning(
+                        "[yellow]DeepSeek doesnt support images. Disabling screenshots[/]"
+                    )
+                elif self.vision == True: # if vision is enabled, add screenshot to chat history
+                    chat_history = await chat_utils.add_screenshot_image_block(screenshot, chat_history)
 
             if context == "ui_state":
                 try:
                     state = self.tools.get_state()
-                    await ctx.set("ui_state", state["a11y_tree"])
+                    await ctx.store.set("ui_state", state["a11y_tree"])
+                    ctx.write_event_to_stream(RecordUIStateEvent(ui_state=state["a11y_tree"]))
                     chat_history = await chat_utils.add_ui_text_block(
                         state["a11y_tree"], chat_history
                     )
@@ -200,11 +206,17 @@ class CodeActAgent(Workflow):
                 success=False, reason="LLM response is None. This is a critical error."
             )
 
+        try:
+            usage = get_usage_from_response(self.llm.class_name(), response)
+        except Exception as e:
+            logger.warning(f"Could not get llm usage from response: {e}")
+            usage = None
+
         await self.chat_memory.aput(response.message)
 
         code, thoughts = chat_utils.extract_code_and_thought(response.message.content)
 
-        event = TaskThinkingEvent(thoughts=thoughts, code=code)
+        event = TaskThinkingEvent(thoughts=thoughts, code=code, usage=usage)
         ctx.write_event_to_stream(event)
         return event
 
@@ -248,7 +260,14 @@ class CodeActAgent(Workflow):
         try:
             self.code_exec_counter += 1
             result = await self.executor.execute(ctx, code)
-            logger.info(f"💡 Code execution successful. Result: {result}")
+            logger.info(f"💡 Code execution successful. Result: {result['output']}")
+            screenshots = result['screenshots']
+            for screenshot in screenshots[:-1]: # the last screenshot will be captured by next step
+                ctx.write_event_to_stream(ScreenshotEvent(screenshot=screenshot))
+
+            ui_states = result['ui_states']
+            for ui_state in ui_states[:-1]:
+                ctx.write_event_to_stream(RecordUIStateEvent(ui_state=ui_state['a11y_tree']))
 
             if self.tools.finished == True:
                 logger.debug("  - Task completed.")
@@ -260,7 +279,7 @@ class CodeActAgent(Workflow):
             
             self.remembered_info = self.tools.memory
             
-            event = TaskExecutionResultEvent(output=str(result))
+            event = TaskExecutionResultEvent(output=str(result['output']))
             ctx.write_event_to_stream(event)
             return event
 
@@ -303,10 +322,11 @@ class CodeActAgent(Workflow):
     async def finalize(self, ev: TaskEndEvent, ctx: Context) -> StopEvent:
         """Finalize the workflow."""
         self.tools.finished = False
-        await ctx.set("chat_memory", self.chat_memory)
+        await ctx.store.set("chat_memory", self.chat_memory)
         
         # Add final state observation to episodic memory
-        await self._add_final_state_observation(ctx)
+        if self.vision:
+            await self._add_final_state_observation(ctx)
         
         result = {}
         result.update(
@@ -329,14 +349,15 @@ class CodeActAgent(Workflow):
         self, ctx: Context, chat_history: List[ChatMessage]
     ) -> ChatResponse | None:
         logger.debug("🔍 Getting LLM response...")
-        messages_to_send = [self.system_prompt] + chat_history
+        limited_history = self._limit_history(chat_history)
+        messages_to_send = [self.system_prompt] + limited_history
         messages_to_send = [chat_utils.message_copy(msg) for msg in messages_to_send]
         try:
             response = await self.llm.achat(messages=messages_to_send)
             logger.debug("🔍 Received LLM response.")
 
             filtered_chat_history = []
-            for msg in chat_history:
+            for msg in limited_history:
                 filtered_msg = chat_utils.message_copy(msg)
                 if hasattr(filtered_msg, "blocks") and filtered_msg.blocks:
                     filtered_msg.blocks = [
@@ -361,9 +382,10 @@ class CodeActAgent(Workflow):
                 chat_history=chat_history_str,
                 response=response_str,
                 timestamp=time.time(),
-                screenshot=(await ctx.get("screenshot", None))
+                screenshot=(await ctx.store.get("screenshot", None))
             )
 
+
             self.episodic_memory.steps.append(step)
 
             assert hasattr(
@@ -385,12 +407,46 @@ class CodeActAgent(Workflow):
                     time.sleep(40)
                 logger.debug("🔍 Retrying call to LLM...")
                 response = await self.llm.achat(messages=messages_to_send)
+            elif (
+                self.llm.class_name() == "Anthropic_LLM"
+                and "overloaded_error" in str(e)
+            ):
+                # Use exponential backoff for Anthropic errors
+                if not hasattr(self, '_anthropic_retry_count'):
+                    self._anthropic_retry_count = 0
+                self._anthropic_retry_count += 1
+                seconds = min(2 ** self._anthropic_retry_count, 60)  # Cap at 60 seconds
+                logger.error(f"Anthropic overload error. Retrying in {seconds} seconds... (attempt {self._anthropic_retry_count})")
+                time.sleep(seconds)
+                logger.debug("🔍 Retrying call to LLM...")
+                response = await self.llm.achat(messages=messages_to_send)
+                self._anthropic_retry_count = 0  # Reset on success
             else:
                 logger.error(f"Could not get an answer from LLM: {repr(e)}")
                 raise e
         logger.debug("  - Received response from LLM.")
         return response
 
+    def _limit_history(
+        self, chat_history: List[ChatMessage]
+    ) -> List[ChatMessage]:
+        if LLM_HISTORY_LIMIT <= 0:
+            return chat_history
+
+        max_messages = LLM_HISTORY_LIMIT * 2
+        if len(chat_history) <= max_messages:
+            return chat_history
+
+        preserved_head: List[ChatMessage] = []
+        if chat_history and chat_history[0].role == "user":
+            preserved_head = [chat_history[0]]
+
+        tail = chat_history[-max_messages:]
+        if preserved_head and preserved_head[0] in tail:
+            preserved_head = []
+
+        return preserved_head + tail
+
     async def _add_final_state_observation(self, ctx: Context) -> None:
         """Add the current UI state and screenshot as the final observation step."""
         try:
