@@ -16,23 +16,28 @@ from droidrun.agent.planner import PlannerAgent
 from droidrun.agent.context.task_manager import TaskManager
 from droidrun.agent.utils.trajectory import Trajectory
 from droidrun.tools import Tools, describe_tools
-from droidrun.agent.common.events import ScreenshotEvent, MacroEvent
+from droidrun.agent.common.events import ScreenshotEvent, MacroEvent, RecordUIStateEvent
 from droidrun.agent.common.default import MockWorkflow
 from droidrun.agent.context import ContextInjectionManager
 from droidrun.agent.context.agent_persona import AgentPersona
 from droidrun.agent.context.personas import DEFAULT
 from droidrun.agent.oneflows.reflector import Reflector
-from droidrun.telemetry import capture, flush, DroidAgentInitEvent, DroidAgentFinalizeEvent
-
+from droidrun.telemetry import (
+    capture,
+    flush,
+    DroidAgentInitEvent,
+    DroidAgentFinalizeEvent,
+)
 
 logger = logging.getLogger("droidrun")
 
+
 class DroidAgent(Workflow):
     """
-A wrapper class that coordinates between PlannerAgent (creates plans) and 
-    CodeActAgent (executes tasks) to achieve a user's goal.
+    A wrapper class that coordinates between PlannerAgent (creates plans) and
+        CodeActAgent (executes tasks) to achieve a user's goal.
     """
-    
+
     @staticmethod
     def _configure_default_logging(debug: bool = False):
         """
@@ -43,20 +48,20 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
         if not logger.handlers:
             # Create a console handler
             handler = logging.StreamHandler()
-            
+
             # Set format
             if debug:
                 formatter = logging.Formatter("%(asctime)s %(levelname)s: %(message)s", "%H:%M:%S")
             else:
                 formatter = logging.Formatter("%(message)s")
-            
+
             handler.setFormatter(formatter)
             logger.addHandler(handler)
             logger.setLevel(logging.DEBUG if debug else logging.INFO)
             logger.propagate = False
-    
+
     def __init__(
-        self, 
+        self,
         goal: str,
         llm: LLM,
         tools: Tools,
@@ -68,35 +73,40 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
         reflection: bool = False,
         enable_tracing: bool = False,
         debug: bool = False,
-        save_trajectories: bool = False,
+        save_trajectories: str = "none",
         excluded_tools: List[str] = None,
         *args,
-        **kwargs
+        **kwargs,
     ):
         """
         Initialize the DroidAgent wrapper.
-        
+
         Args:
             goal: The user's goal or command to execute
             llm: The language model to use for both agents
             max_steps: Maximum number of steps for both agents
             timeout: Timeout for agent execution in seconds
-            reasoning: Whether to use the PlannerAgent for complex reasoning (True) 
+            reasoning: Whether to use the PlannerAgent for complex reasoning (True)
                       or send tasks directly to CodeActAgent (False)
             reflection: Whether to reflect on steps the CodeActAgent did to give the PlannerAgent advice
             enable_tracing: Whether to enable Arize Phoenix tracing
             debug: Whether to enable verbose debug logging
+            save_trajectories: Trajectory saving level. Can be:
+                - "none" (no saving)
+                - "step" (save per step)
+                - "action" (save per action)
             **kwargs: Additional keyword arguments to pass to the agents
         """
         self.user_id = kwargs.pop("user_id", None)
-        super().__init__(timeout=timeout ,*args,**kwargs)
+        super().__init__(timeout=timeout, *args, **kwargs)
         # Configure default logging if not already configured
         self._configure_default_logging(debug=debug)
-        
+
         # Setup global tracing first if enabled
         if enable_tracing:
             try:
                 from llama_index.core import set_global_handler
+
                 set_global_handler("arize_phoenix")
                 logger.info("🔍 Arize Phoenix tracing enabled globally")
             except ImportError:
@@ -114,21 +124,34 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
         self.debug = debug
 
         self.event_counter = 0
-        self.save_trajectories = save_trajectories
-        
+        # Handle backward compatibility: bool -> str mapping
+        if isinstance(save_trajectories, bool):
+            self.save_trajectories = "step" if save_trajectories else "none"
+        else:
+            # Validate string values
+            valid_values = ["none", "step", "action"]
+            if save_trajectories not in valid_values:
+                logger.warning(
+                    f"Invalid save_trajectories value: {save_trajectories}. Using 'none' instead."
+                )
+                self.save_trajectories = "none"
+            else:
+                self.save_trajectories = save_trajectories
+
         self.trajectory = Trajectory(goal=goal)
         self.task_manager = TaskManager()
         self.task_iter = None
 
-        
         self.cim = ContextInjectionManager(personas=personas)
         self.current_episodic_memory = None
 
         logger.info("🤖 Initializing DroidAgent...")
-        
+        logger.info(f"💾 Trajectory saving level: {self.save_trajectories}")
+
         self.tool_list = describe_tools(tools, excluded_tools)
         self.tools_instance = tools
 
+        self.tools_instance.save_trajectories = self.save_trajectories
 
         if self.reasoning:
             logger.info("📝 Initializing Planner Agent...")
@@ -140,14 +163,13 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
                 task_manager=self.task_manager,
                 tools_instance=tools,
                 timeout=timeout,
-                debug=debug
+                debug=debug,
             )
-            self.add_workflows(planner_agent=self.planner_agent)
             self.max_codeact_steps = 5
 
             if self.reflection:
                 self.reflector = Reflector(llm=llm, debug=debug)
-            
+
         else:
             logger.debug("🚫 Planning disabled - will execute tasks directly with CodeActAgent")
             self.planner_agent = None
@@ -167,30 +189,25 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
                 debug=debug,
                 save_trajectories=save_trajectories,
             ),
-            self.user_id
+            self.user_id,
         )
 
-        
         logger.info("✅ DroidAgent initialized successfully.")
 
-    def run(self) -> WorkflowHandler:
+    def run(self, *args, **kwargs) -> WorkflowHandler:
         """
         Run the DroidAgent workflow.
         """
-        return super().run()
-    
+        return super().run(*args, **kwargs)
+
     @step
-    async def execute_task(
-        self,
-        ctx: Context,
-        ev: CodeActExecuteEvent
-        ) -> CodeActResultEvent:
+    async def execute_task(self, ctx: Context, ev: CodeActExecuteEvent) -> CodeActResultEvent:
         """
         Execute a single task using the CodeActAgent.
-        
+
         Args:
             task: Task dictionary with description and status
-            
+
         Returns:
             Tuple of (success, reason)
         """
@@ -215,90 +232,129 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
             handler = codeact_agent.run(
                 input=task.description,
                 remembered_info=self.tools_instance.memory,
-                reflection=reflection
+                reflection=reflection,
             )
-            
+
             async for nested_ev in handler.stream_events():
                 self.handle_stream_event(nested_ev, ctx)
 
             result = await handler
 
-            
             if "success" in result and result["success"]:
-                return CodeActResultEvent(success=True, reason=result["reason"], task=task, steps=result["codeact_steps"])
+                return CodeActResultEvent(
+                    success=True,
+                    reason=result["reason"],
+                    task=task,
+                    steps=result["codeact_steps"],
+                )
             else:
-                return CodeActResultEvent(success=False, reason=result["reason"], task=task, steps=result["codeact_steps"])
-                
+                return CodeActResultEvent(
+                    success=False,
+                    reason=result["reason"],
+                    task=task,
+                    steps=result["codeact_steps"],
+                )
+
         except Exception as e:
             logger.error(f"Error during task execution: {e}")
             if self.debug:
                 import traceback
+
                 logger.error(traceback.format_exc())
             return CodeActResultEvent(success=False, reason=f"Error: {str(e)}", task=task, steps=[])
-    
+
     @step
-    async def handle_codeact_execute(self, ctx: Context, ev: CodeActResultEvent) -> FinalizeEvent | ReflectionEvent:
+    async def handle_codeact_execute(
+        self, ctx: Context, ev: CodeActResultEvent
+    ) -> FinalizeEvent | ReflectionEvent | ReasoningLogicEvent:
         try:
             task = ev.task
             if not self.reasoning:
-                return FinalizeEvent(success=ev.success, reason=ev.reason, output=ev.reason, task=[task], tasks=[task], steps=ev.steps)
-            
-            if self.reflection:
+                return FinalizeEvent(
+                    success=ev.success,
+                    reason=ev.reason,
+                    output=ev.reason,
+                    task=[task],
+                    tasks=[task],
+                    steps=ev.steps,
+                )
+
+            if self.reflection and ev.success:
                 return ReflectionEvent(task=task)
-            
-            return ReasoningLogicEvent()
+
+            # Reasoning is enabled but reflection is disabled.
+            # Success: mark complete and proceed to next step in reasoning loop.
+            # Failure: mark failed and trigger planner immediately without advancing to the next queued task.
+            if ev.success:
+                self.task_manager.complete_task(task, message=ev.reason)
+                return ReasoningLogicEvent()
+            else:
+                self.task_manager.fail_task(task, failure_reason=ev.reason)
+                return ReasoningLogicEvent(force_planning=True)
 
         except Exception as e:
             logger.error(f"❌ Error during DroidAgent execution: {e}")
             if self.debug:
                 import traceback
+
                 logger.error(traceback.format_exc())
             tasks = self.task_manager.get_task_history()
-            return FinalizeEvent(success=False, reason=str(e), output=str(e), task=tasks, tasks=tasks, steps=self.step_counter)
-        
+            return FinalizeEvent(
+                success=False,
+                reason=str(e),
+                output=str(e),
+                task=tasks,
+                tasks=tasks,
+                steps=self.step_counter,
+            )
 
     @step
     async def reflect(
-        self,
-        ctx: Context,
-        ev: ReflectionEvent
-        ) -> ReasoningLogicEvent | CodeActExecuteEvent:
-
-
+        self, ctx: Context, ev: ReflectionEvent
+    ) -> ReasoningLogicEvent | CodeActExecuteEvent:
         task = ev.task
         if ev.task.agent_type == "AppStarterExpert":
             self.task_manager.complete_task(task)
             return ReasoningLogicEvent()
-        
-        reflection = await self.reflector.reflect_on_episodic_memory(episodic_memory=self.current_episodic_memory, goal=task.description)
+
+        reflection = await self.reflector.reflect_on_episodic_memory(
+            episodic_memory=self.current_episodic_memory, goal=task.description
+        )
 
         if reflection.goal_achieved:
             self.task_manager.complete_task(task)
             return ReasoningLogicEvent()
-        
+
         else:
             self.task_manager.fail_task(task)
             return ReasoningLogicEvent(reflection=reflection)
-        
 
     @step
     async def handle_reasoning_logic(
         self,
         ctx: Context,
         ev: ReasoningLogicEvent,
-        planner_agent: Workflow = MockWorkflow()
-        ) -> FinalizeEvent | CodeActExecuteEvent:
+    ) -> FinalizeEvent | CodeActExecuteEvent:
         try:
             if self.step_counter >= self.max_steps:
                 output = f"Reached maximum number of steps ({self.max_steps})"
                 tasks = self.task_manager.get_task_history()
-                return FinalizeEvent(success=False, reason=output, output=output, task=tasks, tasks=tasks, steps=self.step_counter)
+                return FinalizeEvent(
+                    success=False,
+                    reason=output,
+                    output=output,
+                    task=tasks,
+                    tasks=tasks,
+                    steps=self.step_counter,
+                )
             self.step_counter += 1
 
             if ev.reflection:
-                handler = planner_agent.run(remembered_info=self.tools_instance.memory, reflection=ev.reflection)
+                handler = self.planner_agent.run(
+                    remembered_info=self.tools_instance.memory, reflection=ev.reflection
+                )
             else:
-                if self.task_iter:
+                if not ev.force_planning and self.task_iter:
                     try:
                         task = next(self.task_iter)
                         return CodeActExecuteEvent(task=task, reflection=None)
@@ -307,7 +363,9 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
 
                 logger.debug(f"Planning step {self.step_counter}/{self.max_steps}")
 
-                handler = planner_agent.run(remembered_info=self.tools_instance.memory, reflection=None)
+                handler = self.planner_agent.run(
+                    remembered_info=self.tools_instance.memory, reflection=None
+                )
 
             async for nested_ev in handler.stream_events():
                 self.handle_stream_event(nested_ev, ctx)
@@ -320,51 +378,73 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
             if self.task_manager.goal_completed:
                 logger.info(f"✅ Goal completed: {self.task_manager.message}")
                 tasks = self.task_manager.get_task_history()
-                return FinalizeEvent(success=True, reason=self.task_manager.message, output=self.task_manager.message, task=tasks, tasks=tasks, steps=self.step_counter)
+                return FinalizeEvent(
+                    success=True,
+                    reason=self.task_manager.message,
+                    output=self.task_manager.message,
+                    task=tasks,
+                    tasks=tasks,
+                    steps=self.step_counter,
+                )
             if not self.tasks:
                 logger.warning("No tasks generated by planner")
                 output = "Planner did not generate any tasks"
                 tasks = self.task_manager.get_task_history()
-                return FinalizeEvent(success=False, reason=output, output=output, task=tasks, tasks=tasks, steps=self.step_counter)
-            
+                return FinalizeEvent(
+                    success=False,
+                    reason=output,
+                    output=output,
+                    task=tasks,
+                    tasks=tasks,
+                    steps=self.step_counter,
+                )
+
             return CodeActExecuteEvent(task=next(self.task_iter), reflection=None)
-        
+
         except Exception as e:
             logger.error(f"❌ Error during DroidAgent execution: {e}")
             if self.debug:
                 import traceback
+
                 logger.error(traceback.format_exc())
             tasks = self.task_manager.get_task_history()
-            return FinalizeEvent(success=False, reason=str(e), output=str(e), task=tasks, tasks=tasks, steps=self.step_counter)
-    
+            return FinalizeEvent(
+                success=False,
+                reason=str(e),
+                output=str(e),
+                task=tasks,
+                tasks=tasks,
+                steps=self.step_counter,
+            )
 
     @step
-    async def start_handler(self, ctx: Context, ev: StartEvent) -> CodeActExecuteEvent | ReasoningLogicEvent:
+    async def start_handler(
+        self, ctx: Context, ev: StartEvent
+    ) -> CodeActExecuteEvent | ReasoningLogicEvent:
         """
         Main execution loop that coordinates between planning and execution.
-        
+
         Returns:
             Dict containing the execution result
         """
         logger.info(f"🚀 Running DroidAgent to achieve goal: {self.goal}")
         ctx.write_event_to_stream(ev)
-        
+
         self.step_counter = 0
         self.retry_counter = 0
-    
+
         if not self.reasoning:
             logger.info(f"🔄 Direct execution mode - executing goal: {self.goal}")
             task = Task(
                 description=self.goal,
                 status=self.task_manager.STATUS_PENDING,
-                agent_type="Default"
+                agent_type="Default",
             )
-            
+
             return CodeActExecuteEvent(task=task, reflection=None)
-        
+
         return ReasoningLogicEvent()
-                
-        
+
     @step
     async def finalize(self, ctx: Context, ev: FinalizeEvent) -> StopEvent:
         ctx.write_event_to_stream(ev)
@@ -375,7 +455,7 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
                 output=ev.output,
                 steps=ev.steps,
             ),
-            self.user_id
+            self.user_id,
         )
         flush()
 
@@ -387,28 +467,24 @@ A wrapper class that coordinates between PlannerAgent (creates plans) and
             "steps": ev.steps,
         }
 
-        if self.trajectory and self.save_trajectories:
+        if self.trajectory and self.save_trajectories != "none":
             self.trajectory.save_trajectory()
 
         return StopEvent(result)
-    
-    def handle_stream_event(self, ev: Event, ctx: Context):
 
+    def handle_stream_event(self, ev: Event, ctx: Context):
         if isinstance(ev, EpisodicMemoryEvent):
             self.current_episodic_memory = ev.episodic_memory
             return
-        
-  
 
         if not isinstance(ev, StopEvent):
             ctx.write_event_to_stream(ev)
-            
+
             if isinstance(ev, ScreenshotEvent):
                 self.trajectory.screenshots.append(ev.screenshot)
             elif isinstance(ev, MacroEvent):
                 self.trajectory.macro.append(ev)
+            elif isinstance(ev, RecordUIStateEvent):
+                self.trajectory.ui_states.append(ev.ui_state)
             else:
                 self.trajectory.events.append(ev)
-
-
-
