@@ -2,6 +2,7 @@ from llama_index.core.workflow import Event
 from llama_index.core.base.llms.types import ChatMessage
 from typing import Optional, Any
 from droidrun.agent.context import Task
+from droidrun.agent.usage import UsageResult
 
 class PlanInputEvent(Event):
     input: list[ChatMessage]
@@ -10,6 +11,7 @@ class PlanInputEvent(Event):
 class PlanThinkingEvent(Event):
     thoughts: Optional[str] = None
     code: Optional[str] = None  
+    usage: Optional[UsageResult] = None
 
 
 class PlanCreatedEvent(Event):
