@@ -8,7 +8,7 @@ import json
 import time
 import logging
 from llama_index.core.workflow import Context
-from typing_extensions import Optional, Dict, Tuple, List, Any, Type, Self
+from typing import Optional, Dict, Tuple, List, Any
 from droidrun.agent.common.events import (
     InputTextActionEvent,
     KeyPressActionEvent,
@@ -23,13 +23,17 @@ import requests
 import base64
 
 logger = logging.getLogger("droidrun-tools")
+PORTAL_DEFAULT_TCP_PORT = 8080
 
 
 class AdbTools(Tools):
     """Core UI interaction tools for Android device control."""
 
     def __init__(
-        self, serial: str | None = None, use_tcp: bool = False, tcp_port: int = 8080
+        self,
+        serial: str | None = None,
+        use_tcp: bool = False,
+        remote_tcp_port: int = PORTAL_DEFAULT_TCP_PORT,
     ) -> None:
         """Initialize the AdbTools instance.
 
@@ -40,8 +44,7 @@ class AdbTools(Tools):
         """
         self.device = adb.device(serial=serial)
         self.use_tcp = use_tcp
-        self.tcp_port = tcp_port
-        self.tcp_base_url = f"http://localhost:{tcp_port}"
+        self.remote_tcp_port = remote_tcp_port
         self.tcp_forwarded = False
 
         self._ctx = None
@@ -55,11 +58,16 @@ class AdbTools(Tools):
         self.memory: List[str] = []
         # Store all screenshots with timestamps
         self.screenshots: List[Dict[str, Any]] = []
+        # Trajectory saving level
+        self.save_trajectories = "none"
+
+        self.setup_keyboard()
 
         # Set up TCP forwarding if requested
         if self.use_tcp:
             self.setup_tcp_forward()
 
+
     def setup_tcp_forward(self) -> bool:
         """
         Set up ADB TCP port forwarding for communication with the portal app.
@@ -69,18 +77,21 @@ class AdbTools(Tools):
         """
         try:
             logger.debug(
-                f"Setting up TCP port forwarding: tcp:{self.tcp_port} tcp:{self.tcp_port}"
+                f"Setting up TCP port forwarding for port tcp:{self.remote_tcp_port} on device {self.device.serial}"
             )
             # Use adb forward command to set up port forwarding
-            result = self.device.forward(f"tcp:{self.tcp_port}", f"tcp:{self.tcp_port}")
-            self.tcp_forwarded = True
-            logger.debug(f"TCP port forwarding set up successfully: {result}")
+            self.local_tcp_port = self.device.forward_port(self.remote_tcp_port)
+            self.tcp_base_url = f"http://localhost:{self.local_tcp_port}"
+            logger.debug(
+                f"TCP port forwarding set up successfully to {self.tcp_base_url}"
+            )
 
             # Test the connection with a ping
             try:
                 response = requests.get(f"{self.tcp_base_url}/ping", timeout=5)
                 if response.status_code == 200:
                     logger.debug("TCP connection test successful")
+                    self.tcp_forwarded = True
                     return True
                 else:
                     logger.warning(
@@ -105,16 +116,41 @@ class AdbTools(Tools):
         """
         try:
             if self.tcp_forwarded:
-                logger.debug(f"Removing TCP port forwarding for port {self.tcp_port}")
-                result = self.device.forward_remove(f"tcp:{self.tcp_port}")
+                logger.debug(
+                    f"Removing TCP port forwarding for port {self.local_tcp_port}"
+                )
+                # remove forwarding
+                cmd = f"killforward:tcp:{self.local_tcp_port}"
+                logger.debug(f"Removing TCP port forwarding: {cmd}")
+                c = self.device.open_transport(cmd)
+                c.close()
+
                 self.tcp_forwarded = False
-                logger.debug(f"TCP port forwarding removed: {result}")
+                logger.debug(f"TCP port forwarding removed")
                 return True
             return True
         except Exception as e:
             logger.error(f"Failed to remove TCP port forwarding: {e}")
             return False
 
+    def setup_keyboard(self) -> bool:
+        """
+        Set up the DroidRun keyboard as the default input method.
+        Simple setup that just switches to DroidRun keyboard without saving/restoring.
+        
+        Returns:
+            bool: True if setup was successful, False otherwise
+        """
+        try:
+            self.device.shell("ime enable com.droidrun.portal/.DroidrunKeyboardIME")
+            self.device.shell("ime set com.droidrun.portal/.DroidrunKeyboardIME")
+            logger.debug("DroidRun keyboard setup completed")
+            return True
+
+        except Exception as e:
+            logger.error(f"Failed to setup DroidRun keyboard: {e}")
+            return False
+
     def __del__(self):
         """Cleanup when the object is destroyed."""
         if hasattr(self, "tcp_forwarded") and self.tcp_forwarded:
@@ -170,6 +206,7 @@ class AdbTools(Tools):
         except json.JSONDecodeError:
             return None
 
+    @Tools.ui_action
     def tap_by_index(self, index: int) -> str:
         """
         Tap on a UI element by its index.
@@ -328,6 +365,7 @@ class AdbTools(Tools):
         """
         return self.tap_by_index(index)
 
+    @Tools.ui_action
     def swipe(
         self,
         start_x: int,
@@ -372,6 +410,7 @@ class AdbTools(Tools):
             print(f"Error: {str(e)}")
             return False
 
+    @Tools.ui_action
     def drag(
         self, start_x: int, start_y: int, end_x: int, end_y: int, duration: float = 3
     ) -> bool:
@@ -413,6 +452,7 @@ class AdbTools(Tools):
             print(f"Error: {str(e)}")
             return False
 
+    @Tools.ui_action
     def input_text(self, text: str) -> str:
         """
         Input text on the device.
@@ -425,7 +465,6 @@ class AdbTools(Tools):
             Result message
         """
         try:
-            logger.debug(f"Inputting text: {text}")
 
             if self.use_tcp and self.tcp_forwarded:
                 # Use TCP communication
@@ -440,7 +479,7 @@ class AdbTools(Tools):
                 )
 
                 logger.debug(
-                    f"Keyboard input TCP response: {response.status_code}, {response.text}"
+                     f"Keyboard input TCP response: {response.status_code}, {response.text}"
                 )
 
                 if response.status_code != 200:
@@ -448,39 +487,12 @@ class AdbTools(Tools):
 
             else:
                 # Fallback to content provider method
-                # Save the current keyboard
-                original_ime = self.device.shell(
-                    "settings get secure default_input_method"
-                )
-                original_ime = original_ime.strip()
-
-                # Enable the Droidrun keyboard
-                self.device.shell("ime enable com.droidrun.portal/.DroidrunKeyboardIME")
-
-                # Set the Droidrun keyboard as the default
-                self.device.shell("ime set com.droidrun.portal/.DroidrunKeyboardIME")
-
-                # Wait for keyboard to change
-                time.sleep(1)
-
                 # Encode the text to Base64
                 encoded_text = base64.b64encode(text.encode()).decode()
 
                 cmd = f'content insert --uri "content://com.droidrun.portal/keyboard/input" --bind base64_text:s:"{encoded_text}"'
                 self.device.shell(cmd)
 
-                # Wait for text input to complete
-                time.sleep(0.5)
-
-                # Restore the original keyboard
-                if original_ime and "com.droidrun.portal" not in original_ime:
-                    self.device.shell(f"ime set {original_ime}")
-
-                logger.debug(
-                    f"Text input completed: {text[:50]}{'...' if len(text) > 50 else ''}"
-                )
-                return f"Text input completed: {text[:50]}{'...' if len(text) > 50 else ''}"
-
             if self._ctx:
                 input_event = InputTextActionEvent(
                     action_type="input_text",
@@ -501,6 +513,7 @@ class AdbTools(Tools):
         except Exception as e:
             return f"Error sending text input: {str(e)}"
 
+    @Tools.ui_action
     def back(self) -> str:
         """
         Go back on the current view.
@@ -508,13 +521,13 @@ class AdbTools(Tools):
         """
         try:
             logger.debug("Pressing key BACK")
-            self.device.keyevent(3)
+            self.device.keyevent(4)
 
             if self._ctx:
                 key_event = KeyPressActionEvent(
                     action_type="key_press",
                     description=f"Pressed key BACK",
-                    keycode=3,
+                    keycode=4,
                     key_name="BACK",
                 )
                 self._ctx.write_event_to_stream(key_event)
@@ -523,6 +536,7 @@ class AdbTools(Tools):
         except ValueError as e:
             return f"Error: {str(e)}"
 
+    @Tools.ui_action
     def press_key(self, keycode: int) -> str:
         """
         Press a key on the Android device.
@@ -561,6 +575,7 @@ class AdbTools(Tools):
         except ValueError as e:
             return f"Error: {str(e)}"
 
+    @Tools.ui_action
     def start_app(self, package: str, activity: str | None = None) -> str:
         """
         Start an app on the device.
@@ -625,54 +640,47 @@ class AdbTools(Tools):
         except ValueError as e:
             return f"Error: {str(e)}"
 
-    def take_screenshot(self) -> Tuple[str, bytes]:
+    def take_screenshot(self, hide_overlay: bool = True) -> Tuple[str, bytes]:
         """
         Take a screenshot of the device.
         This function captures the current screen and adds the screenshot to context in the next message.
         Also stores the screenshot in the screenshots list with timestamp for later GIF creation.
+        
+        Args:
+            hide_overlay: Whether to hide the overlay elements during screenshot (default: True)
         """
         try:
             logger.debug("Taking screenshot")
+            img_format = "PNG"
+            image_bytes = None
 
             if self.use_tcp and self.tcp_forwarded:
-                # Use TCP communication
-                response = requests.get(f"{self.tcp_base_url}/screenshot", timeout=15)
-
+                # Add hideOverlay parameter to URL
+                url = f"{self.tcp_base_url}/screenshot"
+                if not hide_overlay:
+                    url += "?hideOverlay=false"
+                
+                response = requests.get(url, timeout=10)
                 if response.status_code == 200:
                     tcp_response = response.json()
-
+                    
                     # Check if response has the expected format with data field
-                    if isinstance(tcp_response, dict) and "data" in tcp_response:
+                    if tcp_response.get("status") == "success" and "data" in tcp_response:
+                        # Decode base64 string to bytes
                         base64_data = tcp_response["data"]
-                        try:
-                            # Decode base64 to get image bytes
-                            image_bytes = base64.b64decode(base64_data)
-                            img_format = "PNG"  # Assuming PNG format from TCP endpoint
-                            logger.debug("Screenshot taken via TCP")
-                        except Exception as e:
-                            raise ValueError(
-                                f"Failed to decode base64 screenshot data: {str(e)}"
-                            )
+                        image_bytes = base64.b64decode(base64_data)
+                        logger.debug("Screenshot taken via TCP")
                     else:
-                        # Fallback: assume direct base64 format
-                        try:
-                            image_bytes = base64.b64decode(tcp_response)
-                            img_format = "PNG"
-                            logger.debug("Screenshot taken via TCP (direct base64)")
-                        except Exception as e:
-                            raise ValueError(
-                                f"Failed to decode screenshot response: {str(e)}"
-                            )
+                        # Handle error response from server
+                        error_msg = tcp_response.get("error", "Unknown error")
+                        raise ValueError(f"Error taking screenshot via TCP: {error_msg}")
                 else:
-                    raise ValueError(
-                        f"HTTP request failed with status {response.status_code}: {response.text}"
-                    )
+                    raise ValueError(f"Error taking screenshot via TCP: {response.status_code}")
 
             else:
                 # Fallback to ADB screenshot method
                 img = self.device.screenshot()
                 img_buf = io.BytesIO()
-                img_format = "PNG"
                 img.save(img_buf, format=img_format)
                 image_bytes = img_buf.getvalue()
                 logger.debug("Screenshot taken via ADB")
@@ -710,6 +718,7 @@ class AdbTools(Tools):
         except ValueError as e:
             raise ValueError(f"Error listing packages: {str(e)}")
 
+    @Tools.ui_action
     def complete(self, success: bool, reason: str = ""):
         """
         Mark the task as finished.
@@ -818,14 +827,23 @@ class AdbTools(Tools):
                         "message": "Failed to parse state data from ContentProvider response",
                     }
 
-                if isinstance(state_data, dict) and "data" in state_data:
-                    data_str = state_data["data"]
-                    try:
-                        combined_data = json.loads(data_str)
-                    except json.JSONDecodeError:
+                if isinstance(state_data, dict):
+                    data_str = None
+                    if "data" in state_data:
+                        data_str = state_data["data"]
+                    
+                    if data_str:
+                        try:
+                            combined_data = json.loads(data_str)
+                        except json.JSONDecodeError:
+                            return {
+                                "error": "Parse Error",
+                                "message": "Failed to parse JSON data from ContentProvider response",
+                            }
+                    else:
                         return {
-                            "error": "Parse Error",
-                            "message": "Failed to parse JSON data from ContentProvider data field",
+                            "error": "Format Error", 
+                            "message": "Neither 'data' nor 'message' field found in ContentProvider response",
                         }
                 else:
                     return {
@@ -880,106 +898,6 @@ class AdbTools(Tools):
                 "message": f"Error getting combined state: {str(e)}",
             }
 
-    def get_a11y_tree(self) -> Dict[str, Any]:
-        """
-        Get just the accessibility tree using the /a11y_tree endpoint.
-
-        Returns:
-            Dictionary containing accessibility tree data
-        """
-        try:
-            if self.use_tcp and self.tcp_forwarded:
-                response = requests.get(f"{self.tcp_base_url}/a11y_tree", timeout=10)
-
-                if response.status_code == 200:
-                    tcp_response = response.json()
-
-                    # Check if response has the expected format with data field
-                    if isinstance(tcp_response, dict) and "data" in tcp_response:
-                        data_str = tcp_response["data"]
-                        try:
-                            return json.loads(data_str)
-                        except json.JSONDecodeError:
-                            return {
-                                "error": "Parse Error",
-                                "message": "Failed to parse JSON data from TCP response data field",
-                            }
-                    else:
-                        # Fallback: assume direct JSON format
-                        return tcp_response
-                else:
-                    return {
-                        "error": "HTTP Error",
-                        "message": f"HTTP request failed with status {response.status_code}",
-                    }
-            else:
-                # Fallback: use get_state and extract a11y_tree
-                state = self.get_state()
-                if "error" in state:
-                    return state
-                return {"a11y_tree": state.get("a11y_tree", [])}
-
-        except requests.exceptions.RequestException as e:
-            return {
-                "error": "TCP Error",
-                "message": f"TCP request failed: {str(e)}",
-            }
-        except Exception as e:
-            return {
-                "error": str(e),
-                "message": f"Error getting a11y tree: {str(e)}",
-            }
-
-    def get_phone_state(self) -> Dict[str, Any]:
-        """
-        Get just the phone state using the /phone_state endpoint.
-
-        Returns:
-            Dictionary containing phone state data
-        """
-        try:
-            if self.use_tcp and self.tcp_forwarded:
-                response = requests.get(f"{self.tcp_base_url}/phone_state", timeout=10)
-
-                if response.status_code == 200:
-                    tcp_response = response.json()
-
-                    # Check if response has the expected format with data field
-                    if isinstance(tcp_response, dict) and "data" in tcp_response:
-                        data_str = tcp_response["data"]
-                        try:
-                            return json.loads(data_str)
-                        except json.JSONDecodeError:
-                            return {
-                                "error": "Parse Error",
-                                "message": "Failed to parse JSON data from TCP response data field",
-                            }
-                    else:
-                        # Fallback: assume direct JSON format
-                        return tcp_response
-                else:
-                    return {
-                        "error": "HTTP Error",
-                        "message": f"HTTP request failed with status {response.status_code}",
-                    }
-            else:
-                # Fallback: use get_state and extract phone_state
-                state = self.get_state()
-                if "error" in state:
-                    return state
-                return {"phone_state": state.get("phone_state", {})}
-
-        except requests.exceptions.RequestException as e:
-            return {
-                "error": "TCP Error",
-                "message": f"TCP request failed: {str(e)}",
-            }
-        except Exception as e:
-            return {
-                "error": str(e),
-                "message": f"Error getting phone state: {str(e)}",
-            }
-
     def ping(self) -> Dict[str, Any]:
         """
         Test the TCP connection using the /ping endpoint.
